import hashlib
import os

import requests


def fetch_model(url: str, cache_location: str = "/tmp/zeroshot/cache"):
    """Fetch a model from either a URL or the cache if the file doesn't exist."""
    # We'll hash the file based on the URL
    url_hash = hashlib.sha256(url.encode("utf-8")).hexdigest()
    file_path = os.path.join(cache_location, url_hash)

    # If the file isn't in the cache, download it
    if not os.path.exists(file_path):
        # Make sure the cache directory exists
        os.makedirs(cache_location, exist_ok=True)

        # Download the file
        response = requests.get(url, stream=True)
        response.raise_for_status()

        # Write the file to the cache
        with open(file_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)

    return file_path



import unittest

import numpy as np

from .logistic_regression import LogisticRegression, softmax


class TestLogisticRegression(unittest.TestCase):
    def setUp(self):
        self.logistic_regression = LogisticRegression(
            coefs=np.array([[0.5, -0.5], [0.2, 0.7]]), intercept=np.array([0.1, -0.1])
        )
        self.X = np.array([[0.5, 0.3], [0.7, 0.1], [0.2, 0.6]])

    def test_softmax(self):
        to_test = np.array([[0, 1], [1, 0], [0.2, 0.2], [1, 15]])
        expected_result = np.array(
            [
                [2.68941421e-01, 7.31058579e-01],
                [7.31058579e-01, 2.68941421e-01],
                [5.00000000e-01, 5.00000000e-01],
                [8.31528028e-07, 9.99999168e-01],
            ]
        )

        result = softmax(to_test)
        np.testing.assert_almost_equal(result, expected_result, decimal=7)

    def test_predict_proba(self):
        result = self.logistic_regression.predict_proba(self.X)
        expected_result = np.array(
            [
                [0.49750002, 0.50249998],
                [0.57199613, 0.42800387],
                [0.38698582, 0.61301418],
            ]
        )
        np.testing.assert_almost_equal(result, expected_result, decimal=7)

    def test_predict(self):
        result = self.logistic_regression.predict(self.X)
        expected_result = np.array([1, 0, 1])
        np.testing.assert_array_equal(result, expected_result)


if __name__ == "__main__":
    unittest.main()



import contextlib
import io
import sys

import numpy as np
import requests
from PIL import Image


def _from_data_or_path(input: str | io.BytesIO) -> np.ndarray:
    """Loads the model from either a path or data."""
    img = Image.open(input)
    img = img.convert("RGB")
    img_data = np.array(img)

    return img_data


class nostderr:
    def __init__(self):
        self._original_stderr = None

    def __enter__(self):
        import os

        # Save the original stderr so it can be restored later
        self._original_stderr = sys.stderr

        # Redirect stderr to null device to suppress output
        sys.stderr = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Close the redirected stderr and restore the original one
        sys.stderr.close()
        sys.stderr = self._original_stderr


def numpy_from_path(path: str) -> np.ndarray:
    """Get a numpy array from a path."""
    return _from_data_or_path(path)


def numpy_from_url(path: str) -> np.ndarray:
    """Get a numpy array from a path."""
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0;Win64) AppleWebkit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.82 Safari/537.36"
    }
    response = requests.get(path, headers=headers, timeout=1)
    data = io.BytesIO(response.content)
    return _from_data_or_path(data)



import json
import re
import unittest
import urllib.request
from unittest.mock import mock_open, patch

from .classifier import _infer_path_type, _load_from_file, _load_from_guid


class TestClassifier(unittest.TestCase):
    @patch("urllib.request.urlopen")
    def test_load_from_guid(self, mock_urlopen):
        # Mock API response
        mock_api_response = {"model": "model data"}
        mock_urlopen.return_value.__enter__.return_value.read.return_value = json.dumps(
            mock_api_response
        ).encode()

        guid = "12345678-1234-5678-1234-567812345678"
        expected_result = mock_api_response

        result = _load_from_guid(guid)

        self.assertEqual(result, expected_result)

    @patch("builtins.open", new_callable=mock_open, read_data='{"model": "model data"}')
    def test_load_from_file(self, mock_open):
        # Mock file path
        mock_file_path = "/path/to/model.json"
        with open(mock_file_path) as f:
            expected_result = json.load(f)

        result = _load_from_file(mock_file_path)

        self.assertEqual(result, expected_result)

    def test_infer_path_type(self):
        guid = "12345678-1234-5678-1234-567812345678"
        file_path = "/path/to/model.json"

        self.assertEqual(_infer_path_type(guid), "guid")
        self.assertEqual(_infer_path_type(file_path), "file")


if __name__ == "__main__":
    unittest.main()



import io
import unittest

import numpy as np
from PIL import Image

from .utils import _from_data_or_path


class TestImageLoad(unittest.TestCase):
    def test_all_red_image(self):
        # Create an all-red image of size 10x10
        red_image = Image.new("RGB", (10, 10), "red")

        # Save to a BytesIO object
        img_byte_arr = io.BytesIO()
        red_image.save(img_byte_arr, format="PNG")
        img_byte_arr = io.BytesIO(img_byte_arr.getvalue())

        # Call the function
        result = _from_data_or_path(img_byte_arr)

        # Create the expected output
        expected_output = np.array([[[255, 0, 0]] * 10] * 10)

        # Assert the numpy array is as expected
        np.testing.assert_array_equal(result, expected_output)



import os
import tempfile

import pytest
import requests_mock

from .downloader import fetch_model


def test_fetch_model_downloads_file():
    with tempfile.TemporaryDirectory() as tmpdir:
        url = "http://usezeroshot.com/model"
        expected_content = b"model file"

        with requests_mock.Mocker() as m:
            # Request the file via mocker.
            m.get(url, content=expected_content)

            # Download and check that it worked.
            file_path = fetch_model(url, tmpdir)
            with open(file_path, "rb") as f:
                assert f.read() == expected_content


def test_fetch_model_uses_cache():
    with tempfile.TemporaryDirectory() as tmpdir:
        url = "http://usezeroshot.com/model"
        expected_content = b"model file"

        with requests_mock.Mocker() as m:
            # Request the file via mocker.
            m.get(url, content=expected_content)

            # We'll download twice to check the cache.
            file_path = fetch_model(url, tmpdir)
            file_path_2 = fetch_model(url, tmpdir)

            # Check that the file was not re-downloaded.
            assert m.call_count == 1

            # The file should be the same each time.
            assert file_path == file_path_2



from typing import Callable

import numpy as np
from PIL import Image


def resize(image, new_height, new_width):
    """Resize an image using PIL."""
    img = Image.fromarray(image)
    img = img.resize((new_width, new_height), resample=Image.BICUBIC)
    resized_image = np.array(img)

    return resized_image


def center_crop(image: np.ndarray, new_height: int, new_width: int) -> np.ndarray:
    height, width, _ = image.shape

    # Compute the start coords and crop.
    start_x = (width - new_width) // 2
    start_y = (height - new_height) // 2
    cropped_image = image[start_y : start_y + new_height, start_x : start_x + new_width]

    return cropped_image


def normalize_image(image: np.ndarray) -> np.ndarray:
    if image.dtype == np.uint8:
        image = image.astype(np.float32) / 255.0

    imagenet_mean = np.array([0.485, 0.456, 0.406])
    imagenet_std = np.array([0.229, 0.224, 0.225])

    # Ensure the image has three channels (RGB)
    if image.shape[-1] != 3:
        raise ValueError("Input image must have three channels (RGB)")

    # Normalize the image
    normalized_image = (image - imagenet_mean) / imagenet_std

    return normalized_image.astype(np.float32)


def crop_to_multiple_of_dimension(img: np.ndarray, multiple: int) -> np.ndarray:
    """Crop an image to a multiple of a dimension. Useful for models that require
    input dimensions that are multiples of a certain number, i.e. ViT models.

    Args:
        img: The image to crop.
        multiple: The multiple of each dimension to crop to.

    Returns:
        The cropped image, centered in the original image.
    """
    if multiple <= 0:
        raise ValueError("Multiple must be a positive integer.")

    if len(img.shape) != 3:
        raise ValueError("Unsupported image type, requires 3D array.")

    height, width, _ = img.shape
    if height <= 0 or width <= 0:
        raise ValueError("Image dimensions must be positive integers.")

    # Compute the new dimensions
    new_width = width - (width % multiple)
    new_height = height - (height % multiple)

    # Compute coordinates for the new image
    left = (width - new_width) // 2
    top = (height - new_height) // 2
    right = left + new_width
    bottom = top + new_height

    return img[top:bottom, left:right]


def resize_image_fixed_side(
    image: np.ndarray, side_len: int, method: str = "max"
) -> np.ndarray:
    # Need epsilon for floating point errors.
    epsilon = 1e-3
    height, width, _ = image.shape

    # Compute the scaling factor.
    if method == "max":
        scaling_factor = side_len / max(height, width)
    elif method == "min":
        scaling_factor = side_len / min(height, width)
    else:
        raise ValueError("Invalid method specified. Choose 'max' or 'min'.")

    # Compute the new dimensions plus an epsilon that will get truncated.
    new_height = int(height * scaling_factor + epsilon)
    new_width = int(width * scaling_factor + epsilon)

    assert new_height == side_len or new_width == side_len, "One side incorrect"

    resized_image = resize(image, new_height, new_width)

    return resized_image


def resize_image_min_side(image: np.ndarray, min_side_len: int = 224) -> np.ndarray:
    return resize_image_fixed_side(image, min_side_len, method="min")


def resize_image_max_side(image: np.ndarray, max_side_len: int = 224) -> np.ndarray:
    return resize_image_fixed_side(image, max_side_len, method="max")


def _dino_preprocess(image: np.ndarray) -> np.ndarray:
    """Do standardization and resizing for this model.

    We will do ImageNet standardization, and resize the shortest size to 224, then crop the center 224x224.

    Args:
        image: The image to preprocess.

    Returns:
        np.ndarray: The preprocessed image.
    """
    resized_image = resize_image_min_side(image, 224)
    cropped_image = center_crop(resized_image, 224, 224)

    return normalize_image(cropped_image)


def _dino_large_image(image: np.ndarray) -> np.ndarray:
    """Do standardization and resizing for this model.

    We will do ImageNet standardization, and crop the image to be a multiple of the patch size.

    Args:
        image: The image to preprocess.

    Returns:
        np.ndarray: The preprocessed image.
    """
    cropped_image = crop_to_multiple_of_dimension(image, multiple=14)
    return normalize_image(cropped_image)


def create_preprocess_fn(name: str = "dino") -> Callable[[np.ndarray], np.ndarray]:
    """Creates a preprocessing function from a name."""
    if name == "dino":
        return _dino_preprocess
    elif name == "dino_variable":
        return _dino_large_image
    else:
        raise ValueError(f"Unknown preprocessing function {name}")



import os
import warnings

import numpy as np

from .downloader import fetch_model

_MODELS_ONNX = {
    "dinov2_small": "https://zeroshot-prod-models.s3.us-west-2.amazonaws.com/dinov2_onnx/dinov2_small.onnx",
    "dinov2_base": "",
}

_MODEL_TO_TORCH = {
    "dinov2_small": "dinov2_vits14",
}

_SUPPORTED_BACKENDS = ["onnx", "torch"]


class FeatureExtractor(object):
    def _get_onnx_model(self, name: str):
        import onnxruntime

        # If the model is a file path, just use that model. Otherwise download it.
        if os.path.exists(name):
            self.path = name
        else:
            # Check that the model is valid.
            if name not in _MODELS_ONNX:
                raise ValueError(
                    f"Model {name} not found. Possible values are {_MODELS_ONNX.keys()}"
                )
            self.path = fetch_model(_MODELS_ONNX[name])

        # Load the model via ONNX.
        try:
            with warnings.catch_warnings():
                # Ignore the CUDA warning...
                warnings.filterwarnings("ignore", category=UserWarning)
                self.model = onnxruntime.InferenceSession(
                    self.path, providers=["CUDAExecutionProvider"]
                )
        except:
            self.model = onnxruntime.InferenceSession(self.path)

    def _get_torch_model(self, name: str):
        import torch

        newname = _MODEL_TO_TORCH[name]
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        with warnings.catch_warnings():
            # Ignore the CUDA warning from xformers
            warnings.filterwarnings("ignore", category=UserWarning)
            self.model = torch.hub.load("facebookresearch/dinov2", newname).to(
                self.device
            )

    def _run_onnx_model(
        self, image: np.ndarray, feature_map: bool = False
    ) -> np.ndarray:
        if feature_map:
            raise NotImplementedError("Feature map not supported with ONNX.")

        # Process the model and keep the batch dimension.
        outputs = self.model.run(None, {"input": image})
        return outputs[0]

    def _run_torch_model(self, image: np.ndarray, feature_map=False) -> np.ndarray:
        import torch

        # Get patch sizes.
        _, _, h, w = image.shape
        patches_h, patches_w = h // 14, w // 14

        # Convert to a torch tensor.
        tensor = torch.from_numpy(image).float().to(self.device)

        # Run the model
        with torch.no_grad():
            if feature_map:
                intermediates = self.model.get_intermediate_layers(tensor, n=1)[0]
                np_intermediates = intermediates.cpu().numpy()
                return np.reshape(np_intermediates, (1, patches_h, patches_w, -1))
            else:
                return self.model(tensor).cpu().numpy()

    def __init__(self, name: str, backend="onnx"):
        """Create a feature extractor.

        Args:
            name (str): The name of the model to use, or a path directly to the model.
        """
        self.name = name
        self.path = ""
        self.device = "cpu"
        self.backend = backend

        if self.backend not in _SUPPORTED_BACKENDS:
            raise ValueError(f"Backend must be {', '.join(_SUPPORTED_BACKENDS)}")

        if self.backend == "onnx":
            self._get_onnx_model(name)
        elif self.backend == "torch":
            self._get_torch_model(name)
        else:
            raise ValueError(f"Backend {self.backend} not supported.")

    def process(self, image: np.ndarray, feature_map: bool = False) -> np.ndarray:
        """Process an image into a feature vector.

        Args:
            image (np.ndarray): The image to process, in RGB format with channels last.

        Returns:
            np.ndarray: The feature vector.
        """
        # If the image is 3-dimensional, add a batch dimension.
        if image.ndim == 3:
            image = np.expand_dims(image, axis=0)

        # Make sure the channels is first
        image = np.transpose(image, (0, 3, 1, 2))

        # For now we only support batch size 1, since ONNX doesn't allow variable sizes.
        b, c, h, w = image.shape
        if b != 1:
            raise ValueError("Batch size must be 1 in feature extractor.")

        # Check that the patch sizes are 14
        if h % 14 != 0 or w % 14 != 0:
            raise ValueError(
                "Image size must be divisible by patch size (14) in feature extractor."
            )

        # Ensure we're float32
        image = image.astype(np.float32)

        if self.backend == "onnx":
            return self._run_onnx_model(image, feature_map=feature_map)
        elif self.backend == "torch":
            return self._run_torch_model(image, feature_map=feature_map)
        else:
            raise ValueError(f"Backend {self.backend} not supported.")


class DINOV2FeatureExtractor(FeatureExtractor):
    def __init__(self, size: str = "small", backend: str = "onnx"):
        """Create a DINO v2 feature extractor.

        Args:
            size (str): The size of the model to use, either 'small' or 'base'.
        """
        name = f"dinov2_{size}"
        super().__init__(name, backend=backend)



import numpy as np


def softmax(X: np.ndarray) -> np.ndarray:
    exp_X = np.exp(X - np.max(X, axis=1, keepdims=True))
    return exp_X / np.sum(exp_X, axis=1, keepdims=True)


def expit(X: np.ndarray) -> np.ndarray:
    return 1 / (1 + np.exp(-X))


class LogisticRegression(object):
    def __init__(self, coefs: np.ndarray, intercept: np.ndarray):
        self.coefs = coefs
        self.intercept = intercept

    def predict_proba(self, X: np.ndarray) -> np.ndarray:
        # For a binary classifier, we need to add the negative of the coefficients.
        z = np.dot(X, self.coefs.T) + self.intercept
        if self.coefs.shape[0] == 1:
            prob = expit(z)
            return np.vstack([1 - prob, prob]).T
        else:
            return softmax(z)

    def predict(self, X: np.ndarray) -> np.ndarray:
        probabilities = self.predict_proba(X)
        predictions = np.argmax(probabilities, axis=1)
        return predictions



import unittest

import numpy as np

from .preprocessing import create_preprocess_fn


class TestPreprocessFunctions(unittest.TestCase):
    def test_standard_preprocess_returns_correct_shape_and_range(self):
        preprocess_fn = create_preprocess_fn("dino")
        sample_image = np.random.randint(0, 255, (224, 224, 3), dtype=np.uint8)
        processed_image = preprocess_fn(sample_image)

        self.assertEqual(processed_image.shape, (224, 224, 3))
        self.assertEqual(processed_image.dtype, np.float32)
        self.assertLessEqual(np.mean(processed_image), 1)

    def test_variable_preprocess_returns_correct_shape_and_range(self):
        preprocess_fn = create_preprocess_fn("dino_variable")
        sample_image = np.random.randint(0, 255, (224, 224, 3), dtype=np.uint8)
        processed_image = preprocess_fn(sample_image)

        self.assertEqual(processed_image.shape, (224, 224, 3))
        self.assertEqual(processed_image.dtype, np.float32)
        self.assertLessEqual(np.mean(processed_image), 1)

    def test_variable_preprocess_reduces_shape_to_patch_size(self):
        preprocess_fn = create_preprocess_fn("dino_variable")
        sample_image = np.random.randint(0, 255, (228, 251, 3), dtype=np.uint8)
        processed_image = preprocess_fn(sample_image)

        self.assertEqual(processed_image.shape[0] % 14, 0)
        self.assertEqual(processed_image.shape[1] % 14, 0)

    def test_invalid_preprocess_name_raises(self):
        with self.assertRaises(ValueError):
            create_preprocess_fn("invalid_name")


# This allows running the tests from the command line
if __name__ == "__main__":
    unittest.main()



import os
import unittest
from unittest.mock import patch

import numpy as np
from parameterized import parameterized

from .feature_extractor import DINOV2FeatureExtractor, FeatureExtractor

# Force CPU only for this test
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
os.environ["XFORMERS_DISABLED"] = "1"


class TestFeatureExtractor(unittest.TestCase):
    @patch("os.path.exists")
    def test_invalid_model_name(self, mock_exists):
        """Test if the model name isn't valid."""
        mock_exists.return_value = False

        with self.assertRaises(ValueError):
            FeatureExtractor("invalid_name")

    @patch("os.path.exists")
    def test_invalid_file_path(self, mock_exists):
        mock_exists.return_value = False
        with self.assertRaises(ValueError):
            FeatureExtractor("/invalid/path/to/model")

    @parameterized.expand(
        [
            ((140, 140),),
            ((224, 224),),
            ((448, 448),),
            ((224, 448),),
        ]
    )
    def test_small_dino_single_batch_torch_gives_tensor(
        self, test_shape: tuple[int, int]
    ):
        feature_extractor = DINOV2FeatureExtractor("small", backend="torch")

        # Create a random image that's already normalized
        image = np.random.rand(1, test_shape[0], test_shape[1], 3)
        features = feature_extractor.process(image)

        self.assertSequenceEqual(features.shape, (1, 384))

    def test_small_dino_single_batch_onnx_gives_tensor(self):
        feature_extractor = DINOV2FeatureExtractor("small")

        # Create a random image that's already normalized
        image = np.random.rand(1, 224, 224, 3)
        features = feature_extractor.process(image)

        self.assertSequenceEqual(features.shape, (1, 384))

    def test_small_dino_wrong_shape_onnx_raises(self):
        feature_extractor = DINOV2FeatureExtractor("small")

        image = np.random.rand(1, 448, 448, 3)
        with self.assertRaises(Exception):
            features = feature_extractor.process(image)

    def test_small_dino_wrong_batch_raises(self):
        feature_extractor = DINOV2FeatureExtractor("small")

        image = np.random.rand(4, 224, 224, 3)
        with self.assertRaises(ValueError):
            features = feature_extractor.process(image)


if __name__ == "__main__":
    unittest.main()



import json
import re
import urllib.request
from typing import Any

import numpy as np

from .feature_extractor import DINOV2FeatureExtractor
from .logistic_regression import LogisticRegression
from .preprocessing import create_preprocess_fn
from .utils import numpy_from_path, numpy_from_url

API_ENDPOINT = "https://dvnnfiycsg.execute-api.us-west-2.amazonaws.com/staging"
UUID_PATTERN = r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"


def _load_from_guid(guid: str) -> dict[str, Any]:
    """Loads the model from a guid."""
    # Fetch the model from the API
    fetch_endpoint = f"{API_ENDPOINT}/classifiers/{guid}"
    with urllib.request.urlopen(fetch_endpoint) as response:
        data = json.load(response)
    return data


def _load_from_file(path: str) -> dict[str, Any]:
    """Loads the model from a file."""
    with open(path, "r") as f:
        data = json.load(f)
    return data


def _infer_path_type(path: str) -> str:
    """Infers the path type from the path."""
    uuid_pattern = re.compile(UUID_PATTERN, re.IGNORECASE)
    if uuid_pattern.match(path):
        return "guid"
    else:
        return "file"


class Classifier(object):
    def _load_from_data(self, data: dict) -> None:
        # Load the model
        self.linear_model = LogisticRegression(
            coefs=np.array(data["coefficients"]), intercept=np.array(data["intercepts"])
        )
        self.classes = data["class_list"]
        self.feature_extractor_name = data["feature_extractor"]

        self.feature_extractor = DINOV2FeatureExtractor(
            self.feature_extractor_name, backend=self.backend
        )

    def _image_from_str(self, image: str | np.ndarray) -> np.ndarray:
        """Generate feature vector from a string."""
        if isinstance(image, str) and image.startswith("http"):
            image_np = numpy_from_url(image)
        elif isinstance(image, str):
            image_np = numpy_from_path(image)
        elif isinstance(image, np.ndarray):
            image_np = image

        # Preprocess the image if necessary.
        if self.preprocess_fn is not None:
            image_np = self.preprocess_fn(image_np)

        return image_np

    def __init__(
        self,
        path: str,
        path_type: str = "infer",
        preprocessor: str | None = "dino",
        backend: str = "onnx",
    ):
        # Check that the path type is valid.
        possible_types = ("infer", "guid", "file")
        if path_type not in possible_types:
            raise ValueError(
                f"Path type must be one of {possible_types}, got {path_type}"
            )

        self.path = path
        self.classes = []
        self.backend = backend

        # Load the preprocessor for the model.
        if preprocessor is None:
            self.preprocess_fn = None
        else:
            self.preprocess_fn = create_preprocess_fn(preprocessor)

        # By default we'll just infer the type of path. If it matches a UUID
        # then we'll assume it's a GUID. Since in theory there could be a GUID
        # in the file path, we'll allow the user to override this inference.
        if path_type == "infer":
            path_type = _infer_path_type(path)

        if path_type == "file":
            data = _load_from_file(self.path)
        elif path_type == "guid":
            data = _load_from_guid(self.path)
        self._load_from_data(data)

    def predict(self, image: str | np.ndarray) -> int:
        """Predicts the class of an image.

        Args:
            image: The image to predict, either a url or a numpy array.

        Returns:
            The predicted class
        """
        image = self._image_from_str(image)
        features = self.feature_extractor.process(image)
        return self.linear_model.predict(features)[0]

    def predict_patches(self, image: str | np.ndarray) -> np.ndarray:
        """Predicts the class of an image for each patch

        Args:
            image: The image to predict, either a url or a numpy array.

        Returns:
            The predicted class
        """
        image = self._image_from_str(image)
        features = self.feature_extractor.process(image, feature_map=True)

        # Predict a class for each patch.
        predictions = np.zeros((features.shape[1], features.shape[2]))
        for i in range(features.shape[1]):
            for j in range(features.shape[2]):
                patch = features[:, i, j, :]
                predictions[i, j] = self.linear_model.predict(patch)

        return predictions

    def predict_proba(self, image: str | np.ndarray) -> np.ndarray:
        """Predicts the probabilities of all classes.

        Args:
            image: The image to predict, either a url or a numpy array.

        Returns:
            The predicted class probs.
        """
        image = self._image_from_str(image)
        features = self.feature_extractor.process(image)
        return self.linear_model.predict_proba(features)[0]
